import json
import logging
from time import sleep
import sys
from datetime import datetime
import hashlib
from typing import Union


logger = logging.getLogger(__name__)


def process_browser_log_entry(entry):
    """Принимаем даннаые из performance и возвращаем json"""
    response = json.loads(entry['message'])['message']
    return response


def check_requests(driver):
    # TODO: add FireFox
    """
    Функция смотрит в DevTools Chrome в раздел Network.
    Если у запроса ответ не 200 или 204, кладет в лог урл и статус
    """
    if driver.capabilities["browserName"] == "chrome":
        statuses = [200, 204, 304]
        for performance_data in driver.get_log('performance'):
            requests_json = process_browser_log_entry(performance_data)
            if requests_json["method"] == "Network.responseReceived":
                if requests_json["params"]["response"]["status"] in statuses:
                    pass
                else:
                    logger.error("Request Url: %s, status - %s",
                                 requests_json["params"]["response"]["url"],
                                 requests_json["params"]["response"]["status"])


def notifies_needed():
    """Функция, проверяющая аргументы на наличие --send-notifies (активация отправки уведомлений в ТГ)"""
    if '--send-notifies' in sys.argv:
        logger.info("Notifies are enabled!")
        return True
    else:
        logger.info("Notifies are disabled!")
        return False


def screencast_needed():
    """Функция, проверяющая аргументы на наличие --screencast (запись скринкаста)"""
    if '--screencast' in sys.argv:
        logger.info("Screencast is enabled!")
        return True
    else:
        logger.info("Screencast is disabled!")
        return False


def get_response_data(driver, url: str, wait: int = None):
    # TODO: add FireFox
    """Функция для логгирования ответа от эндпойнта, url которого передаем аргументом"""
    if driver.capabilities["browserName"] == "chrome":
        logger.info("Try to get response data for endpoint %s", url)
        if wait:
            sleep(wait)
        browser_log = driver.get_log('performance')
        events = [process_browser_log_entry(entry) for entry in browser_log]
        events = [event for event in events if 'Network.response' in event['method']]
        tmp_event_list = []
        event_list = []
        for event in events:
            try:
                event_type = event['params']['type']
            except Exception:
                event_type = 'None'
            if event_type == 'XHR':
                tmp_event_list.append(event)
        for event in tmp_event_list:
            try:
                event_url = event['params']['response']['url']
            except Exception:
                event_url = "None"
            if url == event_url:
                event_list.append(event)

        if len(event_list) > 0:
            for i in range(len(event_list)):
                response = driver.execute_cdp_cmd('Network.getResponseBody',
                                                  {'requestId': event_list[i]["params"]["requestId"]})
                logger.info("Endpoint's %s response: %s", url, response['body'])
        else:
            logger.error("Responses list is empty")


def get_datetime():
    now = datetime.now()
    logger.info("Fixed datetime: %s", now)
    return now


def get_timediff(start_time: datetime, end_time: datetime):
    time_diff = end_time - start_time
    logger.info("Time diff: %s", time_diff)
    return time_diff


def get_used_memory(driver):
    # TODO: add FireFox
    """Смотрим в консоли, сколько памяти загрузил в себя js"""
    if driver.capabilities["browserName"] == "chrome":
        return driver.execute_script("""var mem = window.performance.memory.usedJSHeapSize /  1048576; return mem;""")
    else:
        return None


def to_hash(data: str, algorithm: Union[str, bytes]):
    """Вычисление хэша данных с указанием алгоритма при помощи библиотеки hashlib."""
    try:
        hash_generator = getattr(hashlib, algorithm)()
    except AttributeError:
        raise AttributeError("Not supported algorithm. Supported only: sha1, sha256, sha512")

    hash_generator.update(data.encode())
    return hash_generator.hexdigest()



from sqlalchemy import create_engine, MetaData, Table, text
from sqlalchemy.sql import select, update
from sqlalchemy.orm import sessionmaker
import logging

logger = logging.getLogger(__name__)


class QueryMaker:
    """
    Query maker class
    """
    def __init__(self, dburi: str, schema: str):
        self._dburi = dburi
        self._schema = schema
        self._engine = create_engine(self._dburi)
        self._metadata = MetaData()

    @property
    def schema(self):
        return self._schema

    @property
    def engine(self):
        return self._engine

    @property
    def metadata(self):
        return self._metadata

    def make_select_query(self,
                          table_name: str,
                          columns: list = None,
                          limit: int = None,
                          params: str = None):
        """
        Select query generator
        :param table_name: str
        :param schema: str
        :param columns: list
        :param limit: int
        :param params: str
        :return: query
        """
        for _ in range(3):
            try:
                table = Table(table_name, self.metadata, autoload_with=self.engine, schema=self.schema)
                if columns:
                    c_objs = [table.c[col] for col in columns]
                    query = select(*c_objs)
                else:
                    query = select(table)

                if params:
                    query = query.where(text(params))
                if limit:
                    query = query.limit(limit)
                break
            except Exception as query_exception:
                logger.warning("Make select query exception: %s", query_exception)
                continue
        else:
            raise ValueError("Make select query error!")
        return query

    def make_textbased_select_query(self,
                                    text_command: str,
                                    limit: int = None):
        """
        Text based query generator
        :param text_command: str
        :param limit: int
        :return: query
        """
        for _ in range(3):
            try:
                query = select(text(text_command))
                if limit:
                    query = query.limit(limit)
                break
            except Exception as query_exception:
                logger.warning("Make select query exception: %s", query_exception)
                continue
        else:
            raise ValueError("Make select query error!")

        return query

    def make_update_query(self, table_name: str, params: str, values: dict):
        """
        Update query generator
        :param table_name: str
        :param params: str
        :param values: dict
        :return: query
        """
        for _ in range(3):
            try:
                table = Table(table_name, self.metadata, autoload_with=self.engine, schema=self.schema)
                query = update(table).where(text(params)).values(values)
                break
            except Exception as query_exception:
                logger.warning("Make update query exception: %s", query_exception)
                continue
        else:
            raise ValueError("Make update query error!")
        return query

    def run_procedure(self, procedure_name: str, param: str):
        """
        Run procedure method - in work
        :param procedure_name: str
        :param param: str
        :return: list
        """
        connection = self.engine.raw_connection()
        logger.info("created connection: %s", connection)
        try:
            cursor = connection.cursor()
            logger.info("Created cursor: %s", cursor)
            cursor.execute(f"{procedure_name} @{param}")
            results = list(cursor.fetchall())
            cursor.close()
            connection.commit()
        except Exception as e:
            results = None
            logger.error("Run procedure error: %s", e)
        finally:
            connection.close()
        logger.info("Results list: %s", results)
        if results:
            for result in results:
                logger.info("Result: %s", result)


class Executor:
    """Request executor class"""
    def __init__(self, engine):
        self._engine = engine
        self._session = sessionmaker(self._engine)()

    @property
    def session(self):
        return self._session

    def execute(self, query, commit: bool = False):
        """
        execute request method, for update use commit = True
        """
        res = self.session.execute(query)
        if commit is True:
            self.session.commit()
        return res

    def close(self):
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import sys
from pathlib import Path
import os
import shutil


def copy_file(file_name: str):
    base_dir = Path(__file__).parent.resolve()
    user_project_root = os.getcwd()
    source_file = base_dir / file_name
    shutil.copyfile(source_file, user_project_root + f'/{file_name}')


def create_ini_file():
    pytest_ini_path = Path.cwd() / 'pytest.ini'
    with open(pytest_ini_path, 'w') as f:
        f.write('[pytest]\n')
        f.write('addopts = --html=html_reports/report_api_test.html --log-cli-level=INFO --capture=tee-sys  '
                '--self-contained-html\n')


def create_api_project():
    """Create API project structure."""
    project_dir = Path.cwd()
    for folder in ('endpoints', 'tests', 'configs', 'html_reports', 'response_jsons', 'test_data'):
        Path(project_dir / folder).mkdir(parents=True, exist_ok=True)
        open(project_dir / folder / '__init__.py', 'a').close()

    create_ini_file()
    copy_file('conftest.py')


def create_selenium_project():
    """Create Selenium project structure."""
    project_dir = Path.cwd()

    for folder in ('pages', 'processes', 'tests', 'configs', 'html_reports'):
        Path(project_dir / folder).mkdir(exist_ok=True)
        open(project_dir / folder / '__init__.py', 'a').close()

    tmp_dir = Path.cwd() / 'tmp'
    tmp_dir.mkdir(exist_ok=True)
    for subfolder in ('downloads', 'screencasts', 'screenshots'):
        Path(tmp_dir / subfolder).mkdir(parents=True, exist_ok=True)
        open(tmp_dir / subfolder / '__init__.py', 'a').close()

    create_ini_file()
    copy_file('conftest.py')


def main():
    if len(sys.argv) > 1:
        command = sys.argv[1].lower()
        if command == 'create_api_project':
            create_api_project()
        elif command == 'create_selenium_project':
            create_selenium_project()
    else:
        print("Необходимо указать команду для создания проекта.")


if __name__ == '__main__':
    main()



import json
import logging
import requests
from pathlib import Path
from datetime import datetime

logger = logging.getLogger(__name__)


class APIBase:
    """Base class for API"""
    _BASE_DIR = Path(__file__).parent.parent.resolve()
    _headers = {'accept': 'text/plain',
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.90 Safari/537.36',
                'Content-Type': 'application/json-patch+json'}

    @property
    def headers(self):
        return self._headers

    @property
    def base_dir(self):
        return self._BASE_DIR

    def _post(self, url: str, body, return_error: bool = False, return_file: bool = False):
        """POST request impl"""
        response = requests.post(url, data=body, headers=self.headers)
        if response.status_code == 200:
            if return_file is True:
                return response.content
            else:
                return json.loads(response.content)
        else:
            if return_error is True:
                error = self._handle_error(response)
                logger.error("Url: %s POST Error %s: %s", url, response.status_code, error)
                return f'{url} - {response.status_code}: {error}'
            else:
                logger.error("Url: %s POST Error %s: %s", url, response.status_code, response.content)
                raise ValueError(f'{url} - {response.status_code}: {response.content}')

    def _get(self, url: str, return_error: bool = False, return_file: bool = False):
        """GET request impl"""
        response = requests.get(url, headers=self.headers)
        if response.status_code == 200:
            if return_file is True:
                return response.content
            else:
                return json.loads(response.content)
        else:
            if return_error is True:
                error = self._handle_error(response)
                logger.error("Url: %s GET Error %s: %s", url, response.status_code, error)
                return f'{url} - {response.status_code}: {error}'
            else:
                logger.error("Url: %s GET Error %s: %s", url, response.status_code, response.content)
                raise ValueError(f'{url} - {response.status_code}: {response.content}')

    def _delete(self, url: str, data: json = None, return_error: bool = False):
        response = requests.delete(url=url, data=data, headers=self.headers)
        if response.status_code == 200:
            return json.loads(response.content)
        else:
            if return_error is True:
                error = self._handle_error(response)
                logger.error("Url: %s DELETE Error %s: %s", url, response.status_code, error)
                return f'{url} - {response.status_code}: {error}'
            else:
                logger.error("Url: %s DELETE Error %s: %s", url, response.status_code, response.content)
                raise ValueError(f'{url} - {response.status_code}: {response.content}')

    def _put(self, url: str, body: json, return_error: bool = False):
        """PUT request impl"""
        response = requests.put(url, data=body, headers=self.headers)
        if response.status_code == 200:
            return json.loads(response.content)
        else:
            if return_error is True:
                error = self._handle_error(response)
                logger.error("Url: %s PUT Error %s: %s", url, response.status_code, error)
                return f'{url} - {response.status_code}: {error}'
            else:
                logger.error("Url: %s PUT Error %s: %s", url, response.status_code, response.content)
                raise ValueError(f'{url} - {response.status_code}: {response.content}')

    def _send_file(self, url: str, file_name: str, return_error: bool = False):
        """Send file impl"""
        file = Path(f'{self.base_dir}/test_data/{file_name}')
        self._headers['X-File-Size-From'] = '0'
        self._headers['X-File-Size'] = str(file.stat().st_size)
        with open(file, 'rb') as send_file:
            data = send_file.read()
            logger.info("File name: %s", file.name)
            send_data = {'file': (file_name, data, 'multipart/form-data')}
            logger.info("Request headers: %s", self.headers)
            response = requests.post(url, files=send_data, headers=self.headers)
            if response.status_code == 200:
                return json.loads(response.content)
            else:
                if return_error is True:
                    error = self._handle_error(response)
                    logger.error("Url: %s SEND FILE/POST Error %s: %s",
                                 url, response.status_code, error)
                    return f'{url} - {response.status_code}: {error}'
                else:
                    logger.error("Url: %s POST Error %s: %s", url, response.status_code, response.content)
                    raise ValueError(f'{url} - {response.status_code}: {response.content}')

    def _handle_error(self, response):
        """Errors handler"""
        try:
            return json.loads(response.content)
        except json.decoder.JSONDecodeError:
            return response.content

    def _to_json(self, file: str, data: json):
        """Save response to json"""
        with open(f'{self.base_dir}/response_jsons/{file}', 'w+', encoding='utf-8') as file:
            json.dump(data, file, indent=2)

    def _compare_lists(self, list_a: list, list_b: list, equal: bool = False, to_log: bool = True):
        """List comparor"""
        if to_log is True:
            logger.info("Assert data in %s with data in %s", list_a, list_b)
        flag = True
        if len(list_a) != len(list_b):
            flag = False
        else:
            for i in range(len(list_a)):
                if list_a[i] != list_b[i]:
                    flag = False
                    break
        if equal is True:
            assert flag is True, f"Ожидаем, что при ставнении списков flag True, Реально flag {flag}"
        else:
            assert flag is False, f"Ожидаем, что при ставнении списков flag False, Реально flag {flag}"

    def get_now_datetime(self):
        """Get datetime now"""
        return datetime.now()

    def get_timediff(self, t1: datetime, t2: datetime):
        """Get diff between t1 and t2"""
        return int((t2 - t1).total_seconds())

    def auth(self, auth_url: str, email: str, password: str, token_key: str):
        """Auth method"""
        logger.info("Try to auth with acc %s & password %s", email, password)
        body = {"email": email, "password": password}
        response = self._post(auth_url, json.dumps(body))
        self._to_json('auth.json', response)
        logger.info("Token: %s", response[token_key])
        if 'Bearer' in response[token_key]:
            token = response[token_key]
        else:
            token = f"Bearer {response[token_key]}"
        self._headers['Authorization'] = token

    def logout(self):
        """Logout method"""
        try:
            self._headers.pop('Authorization')
            logger.info("Logout success!")
        except KeyError:
            logger.warning("You wasn't logged in!")



from pathlib import Path
from threading import Thread
from typing import Optional
import cv2
import numpy as np
from time import sleep
import logging

logger = logging.getLogger(__name__)


class Recorder:
    def __init__(self, driver, wait_time: float):
        self._driver = driver
        self._wait_time = wait_time  # time between frames
        self._video_th = None
        self._video_process = False
        self.video_file: Optional[Path] = None
        self.BASE_DIR = Path(__file__).parent.resolve()

    def start_video_rec(self, filename):
        logger.info("Start screencast record")
        self._video_process = True
        self.video_file = Path(filename)
        self.video_file.unlink(missing_ok=True)
        self._video_th = Thread(target=self._video_record, name="video_record", daemon=True)
        self._video_th.start()

    def stop_video_rec(self):
        logger.info("Stop screencast record")
        if self._video_th:
            self._video_process = False
            self._video_th.join()

    def _video_record(self):
        def _get_screen():
            self._driver.get_screenshot_as_file(f"{self.BASE_DIR}/tmp/screenshots/last.png")
            return cv2.imread(f"{self.BASE_DIR}/tmp/screenshots/last.png")

        image = _get_screen()
        start_height, start_width, layers = image.shape

        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(str(self.video_file), fourcc, 1.0, (start_width, start_height))
        try:
            while self._video_process:
                image = _get_screen()
                height, width, layers = image.shape
                if (height, width) != (start_height, start_width):
                    blank = np.zeros((start_height, start_width, 3), np.uint8)
                    blank[:, :] = (220, 220, 220)
                    new_img = blank.copy()
                    new_img[0:0 + height, 0:0 + width] = image.copy()
                    image = new_img
                out.write(image)

                sleep(self._wait_time)
        finally:
            out.release()
            cv2.destroyAllWindows()



from setuptools import setup, find_packages


def readme():
    with open('README.md', 'r') as f:
        return f.read()


setup(
    name='OSGRMATT',
    version='0.2.1',
    author='theMizza',
    author_email='dirtymeff@yandex.ru',
    description='Simple framework to fast build selenium (with remote selenium grid as option) and api tests.',
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        'Faker==20.0.3',
        'opencv-python==4.9.0.80',
        'pyodbc==5.0.1',
        'pyTelegramBotAPI==4.14.1',
        'pytest==7.4.3',
        'pytest-html==4.1.1',
        'requests==2.31.0',
        'selenium==4.15.2',
        'webdriver-manager==4.0.1',
        'SQLAlchemy==2.0.23',
        'exchangelib~=5.4.1',
        'numpy~=1.24.4',
        'beautifulsoup4==4.12.3',
        'bs4==0.0.2',
        'python-dotenv~=1.0.1'
    ],
    entry_points={
        'console_scripts': [
            'osgrmatt = OSGRMATT.cli:main'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3.8'
    ],
    keywords='selenium api tests',
    project_urls={
        'GitHub': 'https://github.com/theMizza/OSGRMATT'
    },
    python_requires='>=3.8'
)



from telebot import TeleBot
from pathlib import Path
from datetime import datetime

import logging

logger = logging.getLogger(__name__)


class NotifySender:
    def __init__(self, token: str, chat_id: str):
        self._token = token
        self._chat_id = chat_id
        self._bot = TeleBot(self._token)

    @property
    def bot(self):
        return self._bot

    @property
    def chat_id(self):
        return self._chat_id

    def _send_report_file(self, file_path: Path, caption: str, parse_mode: str = 'html'):
        """
        Send file private method
        :param file_path: Path
        :param caption: str
        :param parse_mode: str
        :return:
        """
        try:
            with open(file_path, 'rb') as document:
                self.bot.send_document(chat_id=self.chat_id, document=document, caption=caption, parse_mode=parse_mode)
                document.close()
            logger.info('Report file was send')
        except Exception as ex:
            logger.error("send_notify_files: %s", ex)

    def send_report_file(self, file_path: Path, caption, parse_mode="html"):
        logger.info("Send report file")
        self._send_report_file(file_path, caption, parse_mode)

    def send_screenshot(self, caption: str, parse_mode: str = 'html', delete_after_send: bool = False):
        """
        Send screenshot to tg
        :param caption: str
        :param parse_mode: str
        :param delete_after_send: bool = False
        :return:
        """
        logger.info("Send screenshot")
        base_dir = Path(__file__).parent.resolve()
        path = Path(base_dir / 'tmp/screenshots/last.png')
        self._send_report_file(path, caption, parse_mode)
        if delete_after_send is True:
            path.unlink()

    def send_screencast(self, caption: str, file_name: str, parse_mode: str = "html", delete_after_send: bool = False):
        """
        Send screencast to tg
        :param file_name: str
        :param caption: str
        :param parse_mode: str
        :param delete_after_send: bool = False
        :return:
        """
        logger.info("Send screencast")
        base_dir = Path(__file__).parent.resolve()
        path = Path(base_dir / f'tmp/screencasts/{file_name}')
        self._send_report_file(path, caption, parse_mode)
        if delete_after_send is True:
            path.unlink()

    def send_start_test_message(self, test_name: str, parse_mode: str = "html"):
        """
        Send start of test message
        :param test_name: str
        :param parse_mode: str
        :return:
        """
        try:
            text = f"================================\n" \
                    f"<b>Test of {test_name} starts!</b>\n" \
                    f"<i>Start time: {datetime.now()}</i>"
            self.bot.send_message(self.chat_id, text=text, parse_mode=parse_mode)
        except Exception as ex:
            logger.error("send_message: %s", ex)

    def send_test_status_message(self, message, parse_mode="html"):
        logger.info("Send text message")
        try:
            self.bot.send_message(chat_id=self.chat_id, text=message, parse_mode=parse_mode)
        except Exception as ex:
            logger.error("send_text_message: %s", ex)



from .api import *
from .browser import *
from .cli import *
from .db import *
from .decorators import *
from .mailer import *
from .notifier import *
from .screen_recorder import *
from .utils import *



import shutil
from typing import Union
from pathlib import Path
from selenium import webdriver
from selenium.common import SessionNotCreatedException
from selenium.webdriver.chrome.options import Options as Chromeoptions
from selenium.webdriver.firefox.options import Options as FireFoxOptions
from selenium.webdriver.remote.file_detector import FileDetector
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager, ChromeType
from time import sleep
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from OSGRMATT.utils import get_used_memory

import logging

logger = logging.getLogger(__name__)
logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)


def get_download_dir():
    """Возвращает путь к папке для скачивания файлов. Не используется."""
    base_dir = Path(__file__).parent.parent.resolve()
    return base_dir / "tmp/downloads"


class LocalFileDetector(FileDetector):
    """Класс отвечает за скачивание и загрузку локальных файлов при использовании ноды грида. Не используется."""
    def is_available(self):
        pass

    def file_detected(self, file_path):
        download_dir = get_download_dir()
        new_file_path = download_dir / Path(file_path).name
        shutil.copyfile(file_path, str(new_file_path))
        logger.info("File saved to path: %s", new_file_path)
        return True

    def is_local_file(self, file_path):
        pass


class Browser:
    def setup_options(self,
                      arguments: list = None,
                      experimental_options: list = None,
                      capabilities: list = None,
                      browser: str = "Chrome"):
        if browser == "FireFox":
            options = FireFoxOptions()
            if experimental_options:
                for key, value in experimental_options[0]['prefs'].items():
                    options.set_preference(key, value)
        else:
            options = Chromeoptions()
            if experimental_options:
                for option in experimental_options:
                    for key, value in option.items():
                        options.add_experimental_option(key, value)
        if arguments:
            for argument in arguments:
                options.add_argument(argument)

        if capabilities:
            for capability in capabilities:
                for key, value in capability.items():
                    options.set_capability(key, value)

        return options

    def setup_driver(self,
                     options: Union['Chromeoptions', 'FireFoxOptions'] = None,
                     command_executor: str = None,
                     browser: str = "Chrome"):
        logger.info("Setup driver")
        if options is None:
            if browser == "FireFox":
                options = FireFoxOptions()
            else:
                options = Chromeoptions()
        if command_executor:
            options.add_argument('--headless')
            options.add_argument('-lang=ru')
            """
            Опция ниже не тестировалась. Подробнее об опции:
            https://www.selenium.dev/documentation/webdriver/drivers/remote_webdriver/#enable-downloads-in-the-grid
            options.enable_downloads = True
            """
            cloud_options = {'build': "build_3",
                             'name': "test_abc3",
                             "browserVersion": "120.0.6099.130"}
            logger.info("Cloud options: %s", cloud_options)
            options.set_capability('cloud:options', cloud_options)
            try:
                driver = webdriver.Remote(command_executor=command_executor,
                                          options=options)
            except SessionNotCreatedException as e:
                logger.error("Create session error: %s", e.msg)
                logger.error("Create session error stacktrace: %s", e.stacktrace)
                raise TypeError("Driver creation error: %s", e.msg)
        else:
            logger.info("Setup driver %s", browser)
            if browser == "FireFox":
                return webdriver.Firefox(options=options)
            else:
                service = Service(ChromeDriverManager(chrome_type=ChromeType.GOOGLE).install())
                driver = webdriver.Chrome(service=service, options=options)

        return driver


class SeleniumLogic:
    def __init__(self, driver):
        self.driver = driver

    def click_element(self, by, value):
        WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((by, value))).click()

    def click_element_fast(self, by, value):
        WebDriverWait(self.driver, 3).until(EC.element_to_be_clickable((by, value))).click()

    def click_element_slow(self, by, value):
        WebDriverWait(self.driver, 180).until(EC.element_to_be_clickable((by, value))).click()

    def click_element_custom(self, by, value, timeout: int):
        WebDriverWait(self.driver, timeout).until(EC.element_to_be_clickable((by, value))).click()

    def enter_text(self, by, value, text, clear: bool = True):
        element = WebDriverWait(self.driver, 30).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        element.send_keys(text)

    def enter_text_fast(self, by, value, text, clear: bool = True):
        element = WebDriverWait(self.driver, 3).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        element.send_keys(text)

    def enter_text_slow(self, by, value, text, clear: bool = True):
        element = WebDriverWait(self.driver, 180).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        element.send_keys(text)

    def enter_text_custom(self, by, value, text, timeout: int, clear: bool = True):
        element = WebDriverWait(self.driver, timeout).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        element.send_keys(text)

    def enter_text_by_letter(self, by, value, text, clear: bool = True):
        element = WebDriverWait(self.driver, 30).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        for i in text:
            element.send_keys(i)
            sleep(0.25)

    def enter_text_by_letter_fast(self, by, value, text, clear: bool = True):
        element = WebDriverWait(self.driver, 3).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        for i in text:
            element.send_keys(i)
            sleep(0.25)

    def enter_text_by_letter_slow(self, by, value, text, clear: bool = True):
        element = WebDriverWait(self.driver, 180).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        for i in text:
            element.send_keys(i)
            sleep(0.25)

    def enter_text_by_letter_custom(self, by, value, text, timeout: int, clear: bool = True):
        element = WebDriverWait(self.driver, timeout).until(EC.visibility_of_element_located((by, value)))
        if clear is True:
            element.clear()
        for i in text:
            element.send_keys(i)
            sleep(0.25)

    def enter_text_in_hidden_input(self, by, value, text):
        """Enter path for upload file"""
        element = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((by, value)))
        element.send_keys(text)

    def enter_text_in_hidden_input_fast(self, by, value, text):
        """Enter path for upload file"""
        element = WebDriverWait(self.driver, 3).until(EC.presence_of_element_located((by, value)))
        element.send_keys(text)

    def enter_text_in_hidden_input_slow(self, by, value, text):
        """Enter path for upload file"""
        element = WebDriverWait(self.driver, 180).until(EC.presence_of_element_located((by, value)))
        element.send_keys(text)

    def enter_text_in_hidden_input_custom(self, by, value, text, timeout: int):
        """Enter path for upload file"""
        element = WebDriverWait(self.driver, timeout).until(EC.presence_of_element_located((by, value)))
        element.send_keys(text)

    def find_element_fast(self, by, value):
        return WebDriverWait(self.driver, 3).until(EC.presence_of_element_located((by, value)))

    def find_element(self, by, value):
        return WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((by, value)))

    def find_element_slow(self, by, value):
        return WebDriverWait(self.driver, 180).until(EC.presence_of_element_located((by, value)))

    def find_element_custom(self, by, value, timeout: int):
        return WebDriverWait(self.driver, timeout).until(EC.presence_of_element_located((by, value)))

    def find_elements(self, by, value):
        """find all presence elements with same locator"""
        return WebDriverWait(self.driver, 30).until(EC.presence_of_all_elements_located((by, value)))

    def find_elements_fast(self, by, value):
        """find all presence elements with same locator"""
        return WebDriverWait(self.driver, 3).until(EC.presence_of_all_elements_located((by, value)))

    def find_elements_slow(self, by, value):
        """find all presence elements with same locator"""
        return WebDriverWait(self.driver, 180).until(EC.presence_of_all_elements_located((by, value)))

    def find_elements_custom(self, by, value, timeout: int):
        """find all presence elements with same locator"""
        return WebDriverWait(self.driver, timeout).until(EC.presence_of_all_elements_located((by, value)))

    def find_visible_elements(self, by, value):
        """find all visible elements with same locator"""
        return WebDriverWait(self.driver, 30).until(EC.visibility_of_all_elements_located((by, value)))

    def find_visible_elements_fast(self, by, value):
        """find all visible elements with same locator"""
        return WebDriverWait(self.driver, 3).until(EC.visibility_of_all_elements_located((by, value)))

    def find_visible_elements_slow(self, by, value):
        """find all visible elements with same locator"""
        return WebDriverWait(self.driver, 180).until(EC.visibility_of_all_elements_located((by, value)))

    def find_visible_elements_custom(self, by, value, timeout: int):
        """find all visible elements with same locator"""
        return WebDriverWait(self.driver, timeout).until(EC.visibility_of_all_elements_located((by, value)))

    def find_element_instantly(self, by, value):
        return self.driver.find_element(by, value)

    def find_elements_instantly(self, by, value):
        return self.driver.find_elements(by, value)

    def check_visible_element(self, by, value):
        return WebDriverWait(self.driver, 30).until(EC.visibility_of_element_located((by, value)))

    def check_visible_fast(self, by, value):
        return WebDriverWait(self.driver, 3).until(EC.visibility_of_element_located((by, value)))

    def check_visible_slow(self, by, value):
        return WebDriverWait(self.driver, 180).until(EC.visibility_of_element_located((by, value)))

    def check_visible_custom(self, by, value, timeout: int):
        return WebDriverWait(self.driver, timeout).until(EC.visibility_of_element_located((by, value)))

    def wait_invisibility_of_element(self, by, value):
        try:
            WebDriverWait(self.driver, 30).until(EC.invisibility_of_element_located((by, value)))
        except TimeoutException:
            return 'need more TO'

    def wait_invisibility_of_element_fast(self, by, value):
        try:
            WebDriverWait(self.driver, 3).until(EC.invisibility_of_element_located((by, value)))
        except TimeoutException:
            return 'need more TO'

    def wait_invisibility_of_element_slow(self, by, value):
        try:
            WebDriverWait(self.driver, 180).until(EC.invisibility_of_element_located((by, value)))
        except TimeoutException:
            return 'need more TO'

    def wait_invisibility_of_element_custom(self, by, value, timeout: int):
        try:
            WebDriverWait(self.driver, timeout).until(EC.invisibility_of_element_located((by, value)))
        except TimeoutException:
            return 'need more TO'

    def send_keys_to_element_instantly(self, by, value, keys):
        element = self.find_elements_instantly(by, value)
        element.send_keys(keys)

    def send_keys_to_element(self, by, value, keys):
        element = self.find_element(by, value)
        element.send_keys(keys)

    def send_keys_to_element_fast(self, by, value, keys):
        element = self.find_element_fast(by, value)
        element.send_keys(keys)

    def send_keys_to_element_slow(self, by, value, keys):
        element = self.find_element_slow(by, value)
        element.send_keys(keys)

    def send_keys_to_element_custom(self, by, value, keys, timeout: int):
        element = self.find_element_custom(by, value, timeout)
        element.send_keys(keys)

    def send_keys_and_press_enter_by_locator(self, locator, keys):
        element = self.find_element(*locator)
        element.send_keys(keys)
        element.send_keys(Keys.ENTER)

    def get_alert_info(self, by, value):
        alert = self.find_element(by, value).text
        return alert

    def get_alert_info_fast(self, by, value):
        alert = self.find_element_fast(by, value).text
        return alert

    def get_alert_info_slow(self, by, value):
        alert = self.find_element_slow(by, value).text
        return alert

    def get_alert_info_custom(self, by, value, timeout: int):
        alert = self.find_element_custom(by, value, timeout).text
        return alert

    def press_escape(self):
        ActionChains(self.driver).send_keys(Keys.ESCAPE).perform()

    def click_element_parametrized(self, contains_text: str):
        dialog_xpath = f"//*[contains(text(), '{contains_text}')]"
        self.click_element(By.XPATH, dialog_xpath)

    def click_element_parametrized_fast(self, contains_text: str):
        dialog_xpath = f"//*[contains(text(), '{contains_text}')]"
        self.click_element_fast(By.XPATH, dialog_xpath)

    def click_element_parametrized_slow(self, contains_text: str):
        dialog_xpath = f"//*[contains(text(), '{contains_text}')]"
        self.click_element_slow(By.XPATH, dialog_xpath)

    def click_element_parametrized_custom(self, contains_text: str, timeout: int):
        dialog_xpath = f"//*[contains(text(), '{contains_text}')]"
        self.click_element_custom(By.XPATH, dialog_xpath, timeout)

    def get_cookies(self):
        return self.driver.get_cookies()

    def get_bearer(self):
        return self.driver.execute_script("return window.localStorage.getItem('token');")

    def ctrl_f5(self):
        ActionChains(self.driver).key_down(Keys.CONTROL).send_keys(Keys.F5).key_up(Keys.CONTROL).perform()

    def get_used_memory(self):
        used_memory = get_used_memory(self.driver)
        logger.info("Used memory in the end: %s Mb", used_memory)
        return str(used_memory)

    def _from_list_of_dicts_to_str(self, data: list):
        string = ''
        for i in data:
            for key, value in i.items():
                string += f'_{key}_{value}_'

        return string

    def check_lists_of_dicts(self, list_a: list, list_b: list):
        first_string = self._from_list_of_dicts_to_str(list_a)
        second_string = self._from_list_of_dicts_to_str(list_b)

        return first_string == second_string

    def download_file_remote(self, downloadable_file: str, download_path: str):
        WebDriverWait(self.driver, 30).until(lambda d: downloadable_file in d.get_downloadable_files())
        files = self.driver.get_downloadable_files()
        logger.info("Files in Selenium Grid: %s", files)
        logger.info("Try to download file from Selenium Grid: %s", downloadable_file)
        try:
            self.driver.download_file(downloadable_file, download_path)
        except EOFError:
            logger.warning("Try to download pdf/zip file issue: https://github.com/SeleniumHQ/selenium/issues/13956")
        self.driver.delete_downloadable_files()

    def tear_down(self):
        self.driver.quit()
